﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Web;
using System.Data;
using System.Data.SQLite;
using System.Text; 


/// <summary>
///MessageManagement 的摘要说明
/// </summary>
public class MessageManagement
{
	static MessageManagement m_Instance = new MessageManagement();

	static public MessageManagement Instance
	{
		get { return m_Instance; }
	}

	private MessageManagement()
	{
	}

	String m_DbPath = "";
	SQLiteConnection m_Conn = null;
	Object m_Lock = new Object();
	Int64 m_MaxKey = 1;
	DateTime m_MaxCreatedTime = DateTime.Now;
	Hashtable m_Listeners = new Hashtable();

	const Int32 MAX_CACHE_COUNT = 4;

	public void Initialize(HttpContext context)
	{
		lock (m_Lock)
		{
			m_DbPath = context.Server.MapPath("~/App_Data/message.db");
			m_Conn = new SQLiteConnection(String.Format("Data Source=\"{0}\";Pooling=False", m_DbPath));
			m_Conn.Open();

			try
			{
				SQLiteCommand cmd = new SQLiteCommand(
					"select max(Key) as MaxKey, max(CreatedTime) as MaxCreatedTime from Message",
					m_Conn
				);

				SQLiteDataReader reader = cmd.ExecuteReader();
				if (reader.Read())
				{
					m_MaxKey = reader[0] == DBNull.Value ? 1 : reader.GetInt64(0);
					m_MaxCreatedTime = reader[1] == DBNull.Value ? DateTime.Now : reader.GetDateTime(1);
				}
			}
			catch
			{
				m_Conn.Close();
				throw;
			}
		}
	}

	public void Dispose()
	{
	}

	public List<Message> Find(String receiver, String sender, Nullable<DateTime> from)
	{
		lock (m_Lock)
		{
			StringBuilder builder = new StringBuilder();
			builder.Append("select Key,Receiver,Sender,Content,CreatedTime from Message where Receiver = ?");
			if (!String.IsNullOrEmpty(sender) || sender == "*") builder.Append(" and Sender = ? ");
			if (from != null) builder.Append(" and CreatedTime > ?");
			builder.Append(" order by CreatedTime asc");

			SQLiteCommand cmd = new SQLiteCommand(builder.ToString(), m_Conn);

			cmd.Parameters.Add("Receiver", DbType.String).Value = receiver;
			if (!String.IsNullOrEmpty(sender) || sender == "*") cmd.Parameters.Add("Sender", DbType.String).Value = sender;
			if (from != null) cmd.Parameters.Add("CreatedTime", DbType.DateTime).Value = from.Value;

			List<Message> messages = new List<Message>();
			SQLiteDataReader reader = cmd.ExecuteReader(CommandBehavior.SingleResult);
			try
			{
				while (reader.Read())
				{
					messages.Add(new Message(reader.GetString(2), reader.GetString(1), reader.GetString(3), reader.GetDateTime(4), reader.GetInt64(0)));
				}
			}
			finally
			{
				reader.Close();
			}
			return messages;
		}
	}

	/// <summary>
	/// 插入新的消息，插入消息后将查询m_Listeners中是否有符合条件的监听器，如存在，同时将消息发送出去
	/// </summary>
	public Message NewMessage(String receiver, String sender, DateTime createdTime, String content)
	{
		lock (m_Lock)
		{
			Message message = new Message(sender, receiver, content, createdTime, ++m_MaxKey);

			List<Message> messages = new List<Message>();
			messages.Add(message);

			if (m_Listeners.ContainsKey(receiver))
			{
				List<MessageListener> listeners = m_Listeners[receiver] as List<MessageListener>;
				List<MessageListener> removeListeners = new List<MessageListener>();
				foreach (MessageListener listener in listeners)
				{
					if ((listener.Sender == "*" || String.Compare(listener.Sender, sender, true) == 0) &&
						(listener.From == null || message.CreatedTime > listener.From))
					{
						listener.Send(messages);
						removeListeners.Add(listener);

						System.Threading.ThreadPool.QueueUserWorkItem(new System.Threading.WaitCallback(listener.Complete));
					}
				}

				foreach (MessageListener listener in removeListeners)
				{
					//移除监听器
					listeners.Remove(listener);
				}
			}

			MessageCacheManagement.Instance.Insert(receiver, message);

			if (MessageCacheManagement.Instance.Count >= MAX_CACHE_COUNT)
			{
				//启动事务
				SQLiteTransaction trans = m_Conn.BeginTransaction();

				try
				{
					List<Message> cacheMsgs = MessageCacheManagement.Instance.GetAll();

					foreach (Message msg in cacheMsgs)
					{
						//超过缓存的最大值，将缓存中的消息全部写入数据库
						SQLiteCommand cmd = new SQLiteCommand(
							"insert into Message (Receiver,Sender,Content,CreatedTime,Key) values (?,?,?,?,?)",
							m_Conn
						);
						cmd.Parameters.Add("Receiver", DbType.String).Value = msg.Receiver;
						cmd.Parameters.Add("Sender", DbType.String).Value = msg.Sender;
						cmd.Parameters.Add("Content", DbType.String).Value = msg.Content;
						cmd.Parameters.Add("CreatedTime", DbType.DateTime).Value = msg.CreatedTime;
						cmd.Parameters.Add("Key", DbType.Int64).Value = msg.Key;

						cmd.ExecuteNonQuery();
					}

					trans.Commit();
				}
				catch
				{
					trans.Rollback();
				}

				MessageCacheManagement.Instance.Clear();
			}

			return message;
		}
	}

	/// <summary>
	/// 添加消息监听器，如果查找到符合监听器条件的消息，返回false，此时不会添加监听器
	/// 如果没有查找到符合监听器条件的消息，返回true，此时监听器将被添加到m_Listeners中
	/// </summary>
	public bool AddListener(String receiver, String sender, Nullable<DateTime> from, WebIM_AsyncResult asynResult)
	{
		MessageListener listener = new MessageListener(receiver, sender, from, asynResult);
		lock (m_Lock)
		{
			if (!m_Listeners.ContainsKey(receiver))
			{
				m_Listeners.Add(receiver, new List<MessageListener>());
			}
			List<MessageListener> listeners = m_Listeners[receiver] as List<MessageListener>;

			//获取用户receiver缓存的消息的最小发送时间
			Nullable<DateTime> min = MessageCacheManagement.Instance.GetMinCreatedTime(receiver);

			List<Message> messages = new List<Message>();

			//当from >= 缓存在内存中的消息的最小时间时，不必查询数据库
			if (min == null || from == null || from.Value < min.Value)
			{
				//查询数据库
				messages.AddRange(Find(receiver, sender, from));
			}

			//在缓存中查询
			messages.AddRange(MessageCacheManagement.Instance.Find(receiver, from.Value));

			if (messages.Count == 0)
			{
				//插入监听器
				listeners.Add(listener);
			}
			else
			{
				//发送消息
				listener.Send(messages);
			}
			return messages.Count == 0;
		}
	}
}

public class MessageCacheManagement
{
	static MessageCacheManagement m_Instance = new MessageCacheManagement();

	static public MessageCacheManagement Instance
	{
		get { return m_Instance; }
	}

	private MessageCacheManagement()
	{
	}

	Int32 m_Count = 0;
	Hashtable m_Cache = new Hashtable();

	List<Message> GetUserMessageCache(String user)
	{
		if (!m_Cache.ContainsKey(user))
		{
			m_Cache.Add(user, new List<Message>());
		}

		return m_Cache[user] as List<Message>;
	}

	/// <summary>
	/// 清除缓存
	/// </summary>
	public void Clear()
	{
		lock (m_Cache)
		{
			List<Message> msgs = new List<Message>();
			foreach (DictionaryEntry ent in m_Cache)
			{
				(ent.Value as List<Message>).Clear();
			}
			m_Count = 0;
		}
	}

	/// <summary>
	/// 获取所有缓存的消息
	/// </summary>
	/// <returns></returns>
	public List<Message> GetAll()
	{
		lock (m_Cache)
		{
			List<Message> msgs = new List<Message>();
			foreach (DictionaryEntry ent in m_Cache)
			{
				foreach (Message msg in ent.Value as List<Message>)
				{
					msgs.Add(msg);
				}
			}
			return msgs;
		}
	}

	/// <summary>
	/// 获取某一用户缓存的消息的最小时间 
	/// </summary>
	public Nullable<DateTime> GetMinCreatedTime(string user)
	{
		lock (m_Cache)
		{
			List<Message> userMsgs = GetUserMessageCache(user);
			return userMsgs.Count == 0 ? null : new Nullable<DateTime>(userMsgs[0].CreatedTime);
		}
	}


	/// <summary>
	/// 在缓存中插入一条消息
	/// </summary>
	/// <param name="user"></param>
	/// <param name="msg"></param>
	public void Insert(String user, Message msg)
	{
		List<Message> userMsgs = null;

		lock (m_Cache)
		{
			userMsgs = GetUserMessageCache(user);
		}

		lock (userMsgs)
		{
			userMsgs.Add(msg);
			m_Count++;
		}
	}

	/// <summary>
	/// 查找缓存中接受者为user，发送时间大于from的消息
	/// </summary>
	public List<Message> Find(String user, DateTime from)
	{
		List<Message> userMsgs = null;

		lock (m_Cache)
		{
			userMsgs = GetUserMessageCache(user);
		}

		lock (userMsgs)
		{
			List<Message> msgs = new List<Message>();

			int i = 0;
			while (i < userMsgs.Count && userMsgs[i].CreatedTime <= from) i++;

			while (i < userMsgs.Count) { msgs.Add(userMsgs[i]); i++; }

			return msgs;
		}
	}

	/// <summary>
	/// 获取消息总量
	/// </summary>
	public Int32 Count
	{
		get { return m_Count; }
	}
}

/// <summary>
/// 消息监听器
/// </summary>
public class MessageListener
{
	public String Receiver, Sender;
	private WebIM_AsyncResult m_AsynResult = null;
	public Nullable<DateTime> From;

	public MessageListener(String receiver, String sender, Nullable<DateTime> from, WebIM_AsyncResult asynResult)
	{
		Receiver = receiver;
		Sender = sender;
		From = from;
		m_AsynResult = asynResult;
	}

	public void Send(List<Message> messages)
	{
		String content = Utility.RenderHashJson(
			"Messages", messages,
			"Receiver", Receiver
		);

		m_AsynResult.Write(content);
	}

	public void Complete(object data)
	{
		m_AsynResult.Complete(data);
	}
}

public class Message : IRenderJson
{
	public String Sender, Receiver, Content;
	public DateTime CreatedTime;
	public Int64 Key;

	public Message(String sender, String receiver, String content, DateTime cteatedTime, Int64 key)
	{
		Sender = sender;
		Receiver = receiver;
		Content = content;
		CreatedTime = cteatedTime;
		Key = key;
	}

	void IRenderJson.RenderJson(StringBuilder builder)
	{
		Utility.RenderHashJson(
			builder,
			"Sender", Sender,
			"Receiver", Receiver,
			"CreatedTime", CreatedTime,
			"Key", Key,
			"Content", Content
		);
	}
}